/**
  *  \file client/widgets/msgaction.h
  *  \brief WMessageActionPanel widget
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_CLIENT_WIDGETS_MSGACTION_H
#define PCC_V2_CLIENT_WIDGETS_MSGACTION_H

#include "client/screens.h"
#include "cpluslib/sigslot.h"
#include "game/coord.h"
#include "game/msg.h"
#include "ui/ui.h"
#include "ui/widgets/button.h"

/** Message action panel. Displays a list of options for an incoming
    message. Options provided are
    - browsing (MD)
    - go to object (*) (MAP)
    - go to X/Y (MAP)
    - reply (*) (MAP)
    - redirect/edit/delete (*)
    - search (MD)
    - write (MD)
    - forward (MD)

    Availability of options marked (*) is controlled by a derived
    class's updateMessageControls() function, which calls setHaveXxx
    for available options.

    Options marked (MAP) are implemented in WMessageActionPanel.
    Options marked (MD) are implemented in WMessageDisplay.
    Redirect/edit/delete must be implemented by a possible user.

    This contains a set of child widgets, which are dynamically
    repositioned and shown or hidden. */
class WMessageActionPanel : public UIWidget {
    GMessagebox& mbox;
    SignalConnection conn_browse;

    UIButton btn_prev;
    UIButton btn_next;
    UIButton btn_goto;
    UIButton btn_reply;
    UIButton btn_xy;
    UIButton btn_redirect;
    UIButton btn_edit;
    UIButton btn_delete;
    UIButton btn_confirm;

    UIButton btn_search;
    UIButton btn_write;
    UIButton btn_forward;

    bool have_goto;
    bool have_reply;
    bool have_xy;
    bool have_outgoing;         // redirect, edit, delete
    bool have_confirm;
    bool hide_goto;

    WScreenId goto_obj;
    int reply_race;
    GPoint xy_pos;

    void showWidget(UIBaseWidget& w, GfxRect r);
    void hideWidget(UIBaseWidget& w);
    void updateWidgets();

 protected:
    void onBrowse();

    void setHaveObject(bool have, WScreenId obj);
    void setHaveReply(bool have, int race);
    void setHaveOutgoing(bool have);
    void setHaveConfirm(bool have);

    virtual void updateMessageControls() = 0;

 public:
    WMessageActionPanel(GMessagebox& mbox);
    ~WMessageActionPanel();
    void drawContent(GfxCanvas& can);
    void getLayoutInfo(LayoutInfo& info);
    void onResize(GfxRect oldsize, bool by_layout);
    bool handleEvent(const UIEvent& event, bool second_pass);

    bool haveObject() const;
    bool haveReply() const;
    bool haveXY() const;

    WScreenId getObject() const;
    int getReply() const;
    GPoint getXY() const;

    void doGoToXY();
    void doGoToObject();
    void doReply();
};

bool findXY(const string_t& str, string_t::size_type n, string_t::size_type& start, string_t::size_type& end, GPoint& pt);


/******************************** Inlines ********************************/

/** Set availability of 'go to' option.
    \param have true iff option is available
    \param obj Object */
inline void
WMessageActionPanel::setHaveObject(bool have, WScreenId obj)
{
    this->have_goto = have;
    this->goto_obj  = obj;
}

/** Set availability of 'reply' option. */
inline void
WMessageActionPanel::setHaveReply(bool have, int race)
{
    this->have_reply = have;
    this->reply_race = race;
}

/** Set availability of outgoing message controls. */
inline void
WMessageActionPanel::setHaveOutgoing(bool have)
{
    this->have_outgoing = have;
}

/** Set availability of Confirm button. */
inline void
WMessageActionPanel::setHaveConfirm(bool have)
{
    this->have_confirm = have;
}

/** Check whether there is a 'go to' option. */
inline bool
WMessageActionPanel::haveObject() const
{
    return have_goto;
}

/** Check whether there is a 'reply' option. */
inline bool
WMessageActionPanel::haveReply() const
{
    return have_reply;
}

/** Check whether there is a 'go to X/Y' option. */
inline bool
WMessageActionPanel::haveXY() const
{
    return have_xy;
}

/** Get object for 'go to' option. \pre haveObject */
inline WScreenId
WMessageActionPanel::getObject() const
{
    return goto_obj;
}

/** Get race for 'reply' option. \pre haveReply */
inline int
WMessageActionPanel::getReply() const
{
    return reply_race;
}

/** Get location for 'go to X/Y' option. \pre haveXY */
inline GPoint
WMessageActionPanel::getXY() const
{
    return xy_pos;
}

#endif
