/**
  *  \file game/formula.h
  *  \brief Math support
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_FORMULA_H
#define PCC_V2_GAME_FORMULA_H

#include <cmath>
#include "cpluslib/types.h"

#undef PI
extern const double PI;

/** Divide a/b, and round result arithmetically. Uses integer
    arithmetic only.
    \pre a >= 0, b > 0 */
inline int32
radiv(int32 a, int32 b)
{
    return (a + (b/2)) / b;
}

/** Compute a/b+plus using IEEE rounding. The result is rounded to the
    nearest integer. If the fractional part is exactly 0.5, rounds to
    the nearest even integer. Uses only integer math. This is hard to
    implement efficiently in C++ with floating point math, anyway.
    \pre a >= 0, b > 0 */
inline long
rdivaddl(long a, int b, int plus)
{
    long x = a / b + plus;
    int r = a % b;
    if (r*2 + (x&1) > b)
        ++x;
    return x;
}

/** Compute a/b+plus using IEEE rounding. \see rdivaddl */
inline int
rdivaddw(int a, int b, int plus)
{
    long x = a / b + plus;
    int r = a % b;
    if (r*2 + (x&1) > b)
        ++x;
    return x;
}

/** Compute square of a number. Provide versions for all integer
    types for best performance and better type checking. */
inline long          square(int a)            { return long(a) * a; }
inline long          square(long a)           { return a * a; }
inline long          square(short a)          { return long(a) * a; }
inline unsigned long square(unsigned int a)   { return static_cast<unsigned long>(a) * a; }
inline unsigned long square(unsigned long a)  { return a * a; }
inline unsigned long square(unsigned short a) { return static_cast<unsigned long>(a) * a; }

inline int32_t dist2FromDX(int dx, int dy)
{
    return square(dx) + square(dy);
}

inline double distFromDX(int dx, int dy)
{
    return std::sqrt(double(dist2FromDX(dx, dy)));
}

int32 roundToInt(double d);

double getHeadingRad(double dx, double dy);

/** Get heading in degrees. */
inline double getHeadingDeg(double dx, double dy)
{
    return getHeadingRad(dx, dy) * (180 / PI);
}

/** Sign of a value.
    \returns -1 if t<0, +1 if t>0, 0 if t==0. */
template<typename T>
inline int
sgn(T t)
{
    return t < 0 ? -1 : t > 0 ? 1 : 0;
}

#endif
