/**
  *  \file game/host.h
  *  \brief Knowledge about Host Properties
  *
  *  Essentially, we have two families of hosts, that's Tim's Host and
  *  PHost. In addition, there are version dependencies (like: Tim's
  *  Host in version 3.22.031+ does not allow you to place high-tech
  *  torps on a low-tech base; or, PHost introduced death rays in
  *  version 4.0.). This module should contain all such host version
  *  dependencies.
  *
  *  As a rule, if the version dependency is "in X, the rule is this,
  *  and in Y, the rule is that", we provide just a function to
  *  distinguish the X and Y cases. In some cases, we also provide a
  *  more detailed function, like "what is the limit for Z in this
  *  host version?".
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_HOST_H
#define PCC_V2_GAME_HOST_H

#include "cpluslib/types.h"

/** Host Type. */
enum GHostType {
    ht_Unknown,                 ///< Not yet determined (must be first).
    ht_Host,                    ///< Tim-Host, any version.
    ht_SRace,                   ///< SRace, any version.
    ht_PHost                    ///< PHost, any version.
};

/** Make version code from three numbers. For example,
    MKVERSION(3,22,31) yiels 322031, the version code for
    THost 3.22.031. If the final component is a letter,
    use its position in the alphabet, e.g. MKVERSION(3,2,5)
    for 3.2e. */
#define MKVERSION(x,y,z) ((x)*100000+(y)*1000+(z))

/** Knowledge about Host properties. This structure provides a
    namespace for all host-specific knowledge required in our client. */
struct GHost {
    /** Host type. */
    GHostType host_type;
    /** Host version. Encoded in the same way as it's visible to scripts:
        322040 = 3.22.40. @see MKVERSION(). */
    long host_version;
    /** Size of HConfig. If we don't get a better information, we use this
        to estimate the host version. */
    int hconfig_size;

    /** Alliance support level for THost. */
    enum TimAllianceSupportLevel {
        ta_NoAlliances,              ///< No alliance support at all.
        ta_SimpleAlliances,          ///< "ff" alliances only.
        ta_StrongAlliances           ///< "FF" alliances (two levels).
    };

    /** Declare that we have a HConfig. */
    inline void setHaveHConfig(int size)
        {
            if (!host_type)
                host_type = ht_Host;
            hconfig_size = size;
        }
    /** Declare that we have a PConfig. */
    inline void setHavePConfig()
        {
            if (!host_type)
                host_type = ht_PHost;
        }

    /** Set host version. Use this only when you know the version
        exactly; from a message maybe. */
    inline void setHostVersion(long v)
        { host_version = v; }

    /** Set host version and type. */
    inline void setHostVersion(long v, GHostType type)
        { host_version = v; host_type = type; }

    /** Check whether this host is the 'PHost' family. */
    inline bool isPHost() const
        { return host_type == ht_PHost; }

    /** Check whether this host is the 'Tim-Host' family. SRace is
        reasonably similar to Tim-Host so it is returned as
        Tim-Host. */
    inline bool isTHost() const
        { return host_type != ht_PHost; }

    /** Check whether this host is SRace. */
    inline bool isSRace() const
        { return host_type == ht_SRace; }

    void finish();
    void clear();

    /** Get ship command argument limit. Many versions have this
        limited to 500, PHost 3.3b raised it to 10000. */
    inline long getCommandArgumentLimit() const
        {
            if (isPHost() && host_version >= MKVERSION(3,3,2))
                return 10000;
            else
                return 500;
        }

    /** Check whether this host version has Death Rays. */
    inline bool hasDeathRays() const
        { return (isPHost() && host_version >= MKVERSION(4,0,0)); }

    /** Check whether this host has ship-specific hull functions. */
    inline bool hasShipSpecificFunctions() const
        { return (isPHost() && host_version >= MKVERSION(4,0,0)); }

    /** Check whether hullfunc.txt assignments are cumulative in this
        host version. This applies to PHost 4.0i/3.4k and later. */
    inline bool hasCumulativeHullfunc() const
        {
            return (isPHost()
                    && (host_version >= MKVERSION(4,0,9)
                        || (host_version < MKVERSION(4,0,0)
                            && host_version >= MKVERSION(3,4,11))));
        }

    /** Check whether ImperialAssault implies PlanetImmunity ability.
        This applies to all hosts except for PHost 4.0i and higher,
        where PlanetImmunity is a separate ability. */
    inline bool hasImmuneAssaultShip() const
        { return (!isPHost() || host_version < MKVERSION(4,0,9)); }

    /** Check whether this host has restrictions in loading high-tech
        torps onto low-tech bases. This is a bug introduced in
        3.22.031, and - given the time Tim spends on Host 3.x - is
        expected to stay for a while. These new hosts do not allow
        you to unload torpedoes onto a starbase if the base doesn't have
        enough tech to build these torps. */
    inline bool hasHighTechTorpedoBug() const
        { return (isTHost() && host_version >= MKVERSION(3,22,31)); }

    /** Check whether siliconoid natives have desert advantage in this
        host. This applies to Tim-Host, as well as PHost 3.3c+. */
    inline bool hasSiliconoidDesertAdvantage() const
        { return (isTHost() || (isPHost() && host_version >= MKVERSION(3,3,3))); }

    /** Check whether this host allows large cargo transfers. A large
        cargo transfer is one where the cargo amount in transit plus
        the ship content exceed the cargo room size. This is allowed
        by most Host versions, but was blocked/made configurable in
        recent THosts. */
    inline bool hasLargeCargoTransfer() const
        { return (isPHost() || (isTHost() && host_version <= MKVERSION(3,22,30))); }

    /** Check THost alliance support. This function is only applicable
        to Tim's host, and returns the support level.
        - 0 if no alliances supported
        - 1 if normal 'ff' allies
        - 2 if strong 'FF' allies */
    inline TimAllianceSupportLevel getTimAllianceLevel() const
        {
            // assume best possible level if host version not known
            if (host_version == 0 || host_version >= MKVERSION(3,22,39))
                return ta_StrongAlliances;
            else if (host_version >= MKVERSION(3,22,7))
                return ta_SimpleAlliances;
            else
                return ta_NoAlliances;
        }

    /** Check whether the "Lay mines in" mission automatically fills
        in the minefield owner. When enlarging a foreing minefield,
        some PHost versions require you to set an 'miX' friendly code,
        some do not. */
    inline bool hasAutomaticMineIdentity() const
        { return (isPHost() && host_version >= MKVERSION(3,4,3)); }

    /** Get post-taxation happiness limit. This is the highest level
        of happiness \em{after taxation happiness effect} at which
        you'll still collect taxes */
    inline int getPostTaxationHappinessLimit() const
        { return isPHost() ? 30 : 31; }

    /** Check whether host allows negative numeric friendly codes.
        This applies to PHost 2.9 and later. */
    inline bool hasNegativeFCodes() const
        { return isPHost() && host_version >= MKVERSION(2,9,0); }

    /** Check whether host allows space-padding in numeric friendly codes.
        This applies to PHost 4.0h or 3.4j and later. */
    inline bool hasSpacePaddedFCodes() const
        { return isPHost() &&
              (host_version >= MKVERSION(4,0,8)
               || (host_version < MKVERSION(4,0,0)
                   && host_version >= MKVERSION(3,4,10))); }

    /** Get the maximum tax for this race. In THost, Borg cannot tax natives
        above 20, Lizards should not tax natives above 75. Note that the 75% rule
        is not enforced by current THost versions due to a bug (it computes
        native/colonist deaths, but forgets to write them back to the data files). */
    int getNativeTaxRateLimit(int player) const;

    /** Get the maximum tax for this race. In THost, Lizards should not tax colonists
        above 75. Note that the 75% rule is not enforced by current THost versions due
        to a bug (it computes native/colonist deaths, but forgets to write them back
        to the data files). */
    int getColonistTaxRateLimit(int player) const;

    /** Check whether the build system of this host has PBP style. */
    bool isPBPGame() const;

    /** Check whether this is a game where ships burn fuel each turn for just being there. */
    bool isEugeneGame() const;

    /** Check whether PHost rounds in mining formulas. This was changed
        in PHost 4.1/3.5 to be more player-friendly. */
    inline bool isPHostRoundingMiningResults() const
        {
            return host_version >= MKVERSION(4,1,0)
                || (host_version < MKVERSION(4,0,0)
                    && host_version >= MKVERSION(3,5,0));
        }

    /** Check for exact hyperjump distance. 340/360 is inclusive in
        PHost, but not in THost. Note that this is actually
        version-dependant, too (old host versions never jump exactly). */
    bool isExactHyperjumpDistance2(int32_t distSquared) const;

    /** Check mission. This allows the host implementation to filter out missions. */
    bool isMissionAllowed(int mission) const;

    /** Check for Minefield-Center bug. If this bug is present, we
        look for the closest minefield center when laying mines,
        regardless whether we actually are in that field. This bug is
        in all THost versions so far. */
    inline bool hasMinefieldCenterBug() const
        {
            return isTHost();
        }

    /** Check for bug in UseAccurateFuelModel computation.
        For PHost with UseAccurateFuelModel enabled, a ship can burn, say, 3.2 kt fuel.
        This means if it starts the move with 4 kt, it will have 1 kt remaining,
        but if it starts the move with 3 kt, it'll run dry. */
    inline bool hasAccurateFuelModelBug() const
        {
            return host_version < MKVERSION(3,4,8)
                || (host_version >= MKVERSION(4,0,0)
                    && host_version < MKVERSION(4,0,5));
        }

    /** Check for alchemy combination function support.
        In PHost 4.0i/3.4k, combinations of alchemy functions produce sensible "best of" functionality,
        e.g. Refinery + AdvancedRefinery = AdvancedRefinery.
        Previously, combinations were not very well thought-out, making the above
        combination produce Refinery only. */
    inline bool hasAlchemyCombinations() const
        {
            return isPHost()
                && (host_version >= MKVERSION(4,0,9)              // 4.0i
                    || (host_version < MKVERSION(4,0,0)
                        && host_version >= MKVERSION(3,4,11)));   // 3.4k
        }

    /** Check for alchemy friendly code support.
        If true, the "alX" and "naX" friendly codes affect refinery ships as well. */
    inline bool hasRefineryFCodes() const
        {
            return isPHost()
                && (host_version >= MKVERSION(4,0,11)             // 4.0k
                    || (host_version < MKVERSION(4,0,0)
                        && host_version >= MKVERSION(3,4,13)));   // 3.4m
        }

    /** Check for alchemy exclusion friendly codes ("naX"). */
    inline bool hasAlchemyExclusionFCodes() const
        { return isPHost(); }

    /** Check for rounding in "alX" alchemy.
        Even with "alX" friendly code, THost consumes supplies in steps of 9. */
    inline bool isAlchemyRounding() const
        { return host_type == ht_Host || host_type == ht_SRace; }

    // /** Check for valid chunnel distance.
    //     Host and PHost implement rules different here.
    //     \param dist2 Distance, squared
    //     \param config Host configuration
    //     \return true if a chunnel with this distance will work */
    // bool isValidChunnelDistance2(int32_t dist2, const game::config::HostConfiguration& config) const;

    /** Get minimum fuel to initiate a chunnel.
        Whereas PHost requires the ship to have remaining fuel after burning 50 kt,
        Host accepts ships that run dry. */
    int getMinimumFuelToInitiateChunnel() const
        { return host_type == ht_PHost ? 51 : 50; }
};
extern GHost host;

#endif
