/**
  *  \file game/rst.cc
  *  \brief Result File Parsing
  *
  *  This module contains GResultFile, a class to read RST file
  *  headers.
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstdlib>
#include <cstring>
#include "game/rst.h"
#include "io/io.h"
#include "util/except.h"
#include "util/nls.h"
#include "cpluslib/assert.h"
#include "cpluslib/format.h"
#include "io/streams.h"

/*! \class GResultFile
    This class parses result files. It provides the user with a means
    of querying header information, such as section positions. */

/** Constructor.
    \param file result file. Must be seekable

    \throws FileFormatException if file is too short or contains an
    invalid pointer. Note that the file is not completely validated,
    just the pointers are checked to be not too badly out of range. 
    User code must handle the case that the file was truncated. */
GResultFile::GResultFile(Stream& file)
    : file(file)
{
    file.seek(0);
    getStructureT(file, header);

    /* Figure out RST version */
    checkHeader();
}

/** Validate header. This also figures out the version number. */
void
GResultFile::checkHeader()
{
    /* Assume standard file */
    version = -1;

    /* RST must be seekable */
    Stream::fpos_t size = file.getSize();
    if (size < 0)
        throw FileFormatException(file, _("Result file is not a regular file"));

    /* Size of Winplan header */
    static const int WINSIZE = 500*8 + 50*12 + 50*4 + 682 + 7800;
    
    if (std::memcmp(header.signature, "VER3.5", 6) == 0
        && header.signature[6] >= '0' && header.signature[6] <= '9'
        && header.signature[7] >= '0' && header.signature[7] <= '9')
    {
        /* may be Winplan RST. Host occasionally sends out RSTs
           bearing the 3.5 header, which are not actually Winplan-style. */
        int32_t kore_ofs = header.offset_win - 1;
        if (kore_ofs >= 0 && kore_ofs <= size - WINSIZE) {
            char buf[4];
            file.seek(kore_ofs + WINSIZE);
            if (file.read(buf, 4) == 4 &&
                (std::memcmp(buf, "1211", 4) == 0 || std::memcmp(buf, "1120", 4) == 0))
            {
                /* It is a Winplan file. */
                version = 10*(header.signature[6]-'0') + header.signature[7]-'0';
            }
        }
    }

    /* Validate pointers. */
    for (int i = 0; i <= sec_LAST; ++i) {
        if (haveSection(Section(i))) {
            int32_t pos = getSectionOffset(Section(i));
            if (pos < 32 || pos >= size) {
                throw FileFormatException(file, format(_("Section %d has an invalid address"), i));
            }
        }
    }
}

/** Get offset of a RST section.
    \param section Section number (sec_xxxx). Only those section present
    in the current type of RST can be queried.
    \returns section offset, suitable for seek.
    \pre haveSection(section) */
int32
GResultFile::getSectionOffset(Section section) const
{
    if (section >= 0 && section < 8) {
        return header.offsets[section] - 1;
    } else if (section == sec_Kore) {
        ASSERT(version >= 0);
        return header.offset_win - 1;
    } else if (section == sec_Leech) {
        ASSERT(version >= 0);
        return header.offset_leech - 1;
    } else if (section == sec_Skore) {
        ASSERT(version >= 1);
        return header.offset_skore - 1;
    } else {
        FAIL();
        return 0;
    }
}

/** Check whether section is present. \returns true iff given section
    is present, i.e. getSectionOffset() works. */
bool
GResultFile::haveSection(Section section) const
{
    if (section >= 0 && section < 8)
        return true;
    else if (section == sec_Kore)
        return version >= 0;
    else if (section == sec_Leech)
        return version >= 0 && header.offset_leech != 0;
    else if (section == sec_Skore)
        return version >= 1;
    else
        return false;
}
