/**
  *  \file game/sim-result.cc
  *  \brief Simulator Result Evaluation
  *
  *  This module contains classes to evaluate simulator results. It
  *  takes raw simulation results and does some statistical evaluation
  *  and grouping with them.
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "game/sim-result.h"
#include "util/misc.h"
#include "cpluslib/assert.h"
#include "util/console.h"
#include "cpluslib/format.h"
#include "util/nls.h"

/****************************** GSimStatItem *****************************/

/** Add single result value.
    \param value [in] Value
    \param w     [in] Weight of this value. */
void
GSimStatItem::add(int32 value, const GSimBattleResult& w)
{
    if (w.this_battle_index == 0) {
        min = max = value;
        min_specimen = max_specimen = w.battles;
    } else {
        if (min > value) {
            min = value;
            min_specimen = w.battles;
        }
        if (max < value) {
            max = value;
            max_specimen = w.battles;
        }
    }
    total_scaled += value * w.this_battle_weight;
}

/********************* GSimResultSummary::UnitResult *********************/

/** Change weight of this unit result. Assuming the result so far was
    obtained using \c old_weight, adjusts all counters such that the
    result is appropriate for \c new_weight.
    \param old_weight Old cumulative weight.
    \param new_weight New cumulative weight. */
void
GSimResultSummary::UnitResult::changeWeight(int32 old_weight, int32 new_weight)
{
    torps_fired.changeWeight(old_weight, new_weight);
    fighters_lost.changeWeight(old_weight, new_weight);
    damage_taken.changeWeight(old_weight, new_weight);
    shields_left.changeWeight(old_weight, new_weight);
    crew_left_or_defense_lost.changeWeight(old_weight, new_weight);
    torps_hit.changeWeight(old_weight, new_weight);
    min_fighters_aboard.changeWeight(old_weight, new_weight);
}

/** Add unit result from ship.
    \param old_ship [in] Original ship
    \param new_ship [in] Ship at end of battle
    \param res      [in] Battle result record (needed for this_battle_index, this_battle_weight) */
void
GSimResultSummary::UnitResult::addResult(const GSimShip& old_ship, const GSimShip& new_ship, const GSimBattleResult& res)
{
    /* Overall result */
    if (new_ship.getStat().num_fights != 0)
        ++fought;
    if (new_ship.getOwner() != 0) {
        if (new_ship.getOwner() == old_ship.getOwner())
            ++won;
        else
            ++captured;
    }

    /* Statistics counters */
    // FIXME: NTP?
    if (old_ship.getTorpLauncherCount() != 0)
        torps_fired.add(old_ship.getAmmo() - new_ship.getAmmo(), res);
    else
        torps_fired.add(0, res);

    if (old_ship.getBayCount() != 0)
        fighters_lost.add(old_ship.getAmmo() - new_ship.getAmmo(), res);
    else
        fighters_lost.add(0, res);

    damage_taken.add(new_ship.getDamage(), res);
    shields_left.add(new_ship.getShield(), res);
    crew_left_or_defense_lost.add(new_ship.getCrew(), res);

    if (old_ship.getTorpLauncherCount() != 0)
        torps_hit.add(new_ship.getStat().torps_hit, res);

    if (old_ship.getBayCount() != 0)
        min_fighters_aboard.add(new_ship.getStat().min_fighters_aboard, res);
}

/** Add unit result from planet.
    \param old_planet [in] Original planet
    \param new_planet [in] Planet at end of battle
    \param res        [in] Battle result record (needed for this_battle_index, this_battle_weight) */
void
GSimResultSummary::UnitResult::addResult(const GSimPlanet& old_planet, const GSimPlanet& new_planet, const GSimBattleResult& res)
{
    /* Overall result */
    if (new_planet.getStat().num_fights != 0)
        ++fought;
    if (new_planet.getOwner() != 0) {
        if (new_planet.getOwner() == old_planet.getOwner())
            ++won;
        else
            ++captured;
    }

    /* Statistics counters */
    // FIXME: GSimStatItem torps_fired;
    fighters_lost.add(old_planet.getBaseFighters() - new_planet.getBaseFighters(), res);
    damage_taken.add(new_planet.getDamage(), res);
    shields_left.add(new_planet.getShield(), res);
    crew_left_or_defense_lost.add(old_planet.getDefense() - new_planet.getDefense(), res);

    torps_hit.add(new_planet.getStat().torps_hit, res);
    min_fighters_aboard.add(new_planet.getStat().min_fighters_aboard, res);
}

/********************* GSimResultSummary::ClassResult ********************/

/** Constructor. Builds a new ClassResult from simulator output.
    \param new_ship  [in] Simulator output
    \param result    [in] Simulator meta-information */
GSimResultSummary::ClassResult::ClassResult(const GSimState& new_state, const GSimBattleResult& result)
    : owned_units(),
      cumulative_weight(result.this_battle_weight),
      sample_battle(result.battles)
{
    /* initialize */
    owned_units.setAll(0);

    /* count it */
    for (GSimState::ship_index_t i = 0; i < new_state.getNumShips(); ++i) {
        int o = new_state.getShip(i).getOwner();
        if (owned_units.isValidIndex(o))
            ++owned_units[o];
    }
    if (new_state.hasPlanet()) {
        int o = new_state.getPlanet().getOwner();
        if (owned_units.isValidIndex(o))
            ++owned_units[o];
    }
}

GSimResultSummary::ClassResult::~ClassResult()
{ }

/** Check for equal class.
    \return true iff this class and \c other are the same class */
bool
GSimResultSummary::ClassResult::isSameClass(const ClassResult& other) const
{
    for (int i = 1; i <= 12; ++i)
        if (owned_units[i] != other.owned_units[i])
            return false;
    return true;
}

/** Add new result of same class. If another result of the same class
    as this one has been obtained, this adjusts the statistics counters
    accordingly.
    \pre isSameClass(other) */
void
GSimResultSummary::ClassResult::addNewSameClassResult(const ClassResult& other)
{
    ASSERT(isSameClass(other));

    cumulative_weight += other.cumulative_weight;
    sample_battle = other.sample_battle;
}

/*************************** GSimResultSummary ***************************/

/** Constructor. Make blank result. */
GSimResultSummary::GSimResultSummary()
    : total_weight(0), cumulative_weight(0), battle_count(0), planet_index(-1)
{ }

/** Destructor. */
GSimResultSummary::~GSimResultSummary()
{ }

/** Incorporate result into this object.
    \param old_state [in] Simulator input
    \param new_state [in] Simulator output
    \param result    [in] Result meta-information provided by simulator */
void
GSimResultSummary::addResult(const GSimState& old_state,
                             const GSimState& new_state,
                             GSimBattleResult result)
{
    /* check validity of parameters */
    ASSERT(old_state.getNumShips() == new_state.getNumShips());
    ASSERT(old_state.hasPlanet() == new_state.hasPlanet());

    /* build unit_results on first iteration */
    if (unit_results.empty()) {
        for (GSimState::ship_index_t i = 0; i < old_state.getNumShips(); ++i)
            unit_results.push_back(new UnitResult());
        if (old_state.hasPlanet()) {
            planet_index = unit_results.size();
            unit_results.push_back(new UnitResult());
        }
        total_weight = result.total_battle_weight;
    }

    /* adjust weights */
    if (total_weight < result.total_battle_weight) {
        /* the new battle has a higher weight, so we have to update
           the existing results. This should not happen if the
           simulator works right, but it's not too bad either. */
        console.write(LOG_WARN,
                      format(_("Simulator: changing weight from %d to %d"),
                             total_weight, result.total_battle_weight));
        for (unitvec_t::iterator i = unit_results.begin(); i != unit_results.end(); ++i)
            (*i)->changeWeight(total_weight, result.total_battle_weight);
        for (classvec_t::iterator i = class_results.begin(); i != class_results.end(); ++i)
            (*i)->changeWeight(total_weight, result.total_battle_weight);

        cumulative_weight = 0;
        for (classvec_t::iterator i = class_results.begin(); i != class_results.end(); ++i)
            cumulative_weight += (*i)->cumulative_weight;

        total_weight = result.total_battle_weight;
    }

    if (total_weight > result.total_battle_weight) {
        /* the battle has a lower weight, so raise its weight. This
           should not happen if the simulator works right, but it's
           not too bad either. */
        console.write(LOG_WARN,
                      format(_("Simulator: incoming fight weight %d, expecting %d"),
                             result.total_battle_weight, total_weight));
        result.changeWeightTo(total_weight);
    }
    ASSERT(result.total_battle_weight == total_weight);

    /* Now compute new unit results */
    for (GSimState::ship_index_t i = 0; i < old_state.getNumShips(); ++i) {
        unit_results[i]->addResult(old_state.getShip(i),
                                   new_state.getShip(i),
                                   result);
    }
    if (old_state.hasPlanet()) {
        ASSERT(planet_index >= 0);
        unit_results[planet_index]->addResult(old_state.getPlanet(),
                                              new_state.getPlanet(),
                                              result);
    }

    /* And add it to the class results */
    ClassResult this_class(new_state, result);
    bool have_this_class = false;
    for (classvec_t::iterator i = class_results.begin(); i != class_results.end(); ++i) {
        if ((*i)->isSameClass(this_class)) {
            (*i)->addNewSameClassResult(this_class);
            have_this_class = true;
            updateClassResultSortOrder(i - class_results.begin());
            break;
        }
    }
    if (!have_this_class) {
        class_results.push_back(new ClassResult(this_class));
        updateClassResultSortOrder(class_results.size() - 1);
    }

    /* Finally, adjust our counters */
    cumulative_weight += result.this_battle_weight;
    ++battle_count;
}

/** Update class result sort order. Assuming value at change_index was
    modified (count increased), sort it into its place. */
void
GSimResultSummary::updateClassResultSortOrder(classvec_t::size_type change_index)
{
    while (change_index > 0) {
        if (class_results[change_index-1]->cumulative_weight < class_results[change_index]->cumulative_weight) {
            /* swap them */
            class_results.swap_elems(change_index-1, change_index);
            change_index--;
        } else {
            /* finish */
            break;
        }
    }
}
