/**
  *  \file u/t_arch_directory.cc
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstdio>
#include "io/dirs.h"

#include "u/t_arch.h"
#include "util/except.h"
#include "arch/platform.h"

static const char FILE_NAME1[] = "01file01.dat";
static const char FILE_NAME2[] = "02file02.dat";

ArchDirectoryTestSuite::ArchDirectoryTestSuite()
{
    std::remove(FILE_NAME1);
    std::remove(FILE_NAME2);
}

ArchDirectoryTestSuite::~ArchDirectoryTestSuite()
{ }

void
ArchDirectoryTestSuite::checkDirectoryForFile(Directory& dir, const char* name, bool mustExist)
{
    // Pass 1: try by iterating
    bool seenTheFile = false;
    Ptr<Enumeration<Ptr<DirectoryEntry> > > en = dir.getEntries();
    TS_ASSERT(en != 0);

    while (en->hasMoreElements()) {
        Ptr<DirectoryEntry> entry = en->getNextElement();
        TS_ASSERT(entry != 0);

        // Use case-insensitive compare; the filesystem might not be case-preserving.
        if (strCaseCmp(entry->getBaseName(), name) == 0) {
            seenTheFile = true;
            TS_ASSERT_EQUALS(int(entry->getFileType()), int(DirectoryEntry::FT_FILE));
        }
    }
    TS_ASSERT_EQUALS(seenTheFile, mustExist);

    // Pass 2: try direct lookup
    Ptr<DirectoryEntry> entry = dir.getEntryByName(name);
    TS_ASSERT(entry != 0);
    if (mustExist) {
        TS_ASSERT_EQUALS(int(entry->getFileType()), int(DirectoryEntry::FT_FILE));
    }
}

void
ArchDirectoryTestSuite::setUp()
{
    /* We need the NLS stuff initialized to be able to handle UTF-8
       file names which are tested in testCreationAndDeletion(). */
    Platform::initPlatform();
}

void
ArchDirectoryTestSuite::testDirectoryAccess()
{
    // Open current directory
    string_t s = Directory::getWorkingDirectory();
    TS_ASSERT(Directory::isAbsoluteName(s));

    Ptr<Directory> dir = Directory::openDir(s);
    TS_ASSERT(dir != 0);

    // Create a file
    Ptr<DirectoryEntry> entry = dir->getEntryByName(FILE_NAME1);
    TS_ASSERT(entry != 0);

    {
        Ptr<Stream> s = entry->openFile(Stream::C_CREATE);
        TS_ASSERT(s != 0);
        int written = s->write("foo", 3);
        TS_ASSERT_EQUALS(written, 3);
    }

    // Check that the file is actually there
    checkDirectoryForFile(*dir, FILE_NAME1, true);
    checkDirectoryForFile(*dir, FILE_NAME2, false);

    // Rename the file
    TS_ASSERT_THROWS_NOTHING(entry->renameTo(FILE_NAME2));

    // Directory entry continues to refer to old name, which cannot be opened
    TS_ASSERT_THROWS(entry->openFile(Stream::C_READ), FileOpenException);

    // New entry refers to new file
    entry = dir->getEntryByName(FILE_NAME2);
    TS_ASSERT(entry != 0);
    TS_ASSERT_EQUALS(entry->getSize(), 3);
    TS_ASSERT_EQUALS(int(entry->getFileType()), int(DirectoryEntry::FT_FILE));

    checkDirectoryForFile(*dir, FILE_NAME1, false);
    checkDirectoryForFile(*dir, FILE_NAME2, true);

    // Try reading the file
    {
        Ptr<Stream> s = entry->openFile(Stream::C_READ);
        TS_ASSERT(s != 0);
        char tmp[10];
        int read = s->read(tmp, sizeof(tmp));
        TS_ASSERT_EQUALS(read, 3);
        TS_ASSERT_SAME_DATA(tmp, "foo", 3);
    }

    // Erase the file
    TS_ASSERT_THROWS_NOTHING(entry->erase());

    // Directory entry continues to refer to old name, which cannot be opened
    TS_ASSERT_THROWS(entry->openFile(Stream::C_READ), FileOpenException);

    checkDirectoryForFile(*dir, FILE_NAME1, false);
    checkDirectoryForFile(*dir, FILE_NAME2, false);
}

void
ArchDirectoryTestSuite::testCreationAndDeletion()
{
    Ptr<Directory> dir = Directory::openDir(Directory::getWorkingDirectory());

    // Get some entry
    Ptr<DirectoryEntry> entry;
    TS_ASSERT_THROWS_NOTHING(entry = dir->getEntryByName("03dir"));
    TS_ASSERT(entry != 0);
    TS_ASSERT_THROWS_NOTHING(entry->getFileType());

    // Create it
    TS_ASSERT_THROWS_NOTHING(entry->createAsDirectory());
    TS_ASSERT_EQUALS(int(entry->getFileType()), int(DirectoryEntry::FT_DIRECTORY));

    // Create a file in it
    Ptr<Directory> subdir;
    TS_ASSERT_THROWS_NOTHING(subdir = entry->openDir());
    TS_ASSERT(subdir != 0);
    TS_ASSERT_THROWS_NOTHING(subdir->openFile("test", Stream::C_CREATE));
    TS_ASSERT_THROWS_NOTHING(subdir->erase("test"));

    // Erase directory again
    TS_ASSERT_THROWS_NOTHING(entry->erase());

    // Try same thing with Unicode name
    TS_ASSERT_THROWS_NOTHING(entry = dir->getEntryByName("03b\xc3\xa4r"));
    TS_ASSERT(entry != 0);
    TS_ASSERT_THROWS_NOTHING(entry->getFileType());
    try {
        entry->createAsDirectory();
        TS_ASSERT_EQUALS(int(entry->getFileType()), int(DirectoryEntry::FT_DIRECTORY));
        TS_ASSERT_THROWS_NOTHING(entry->erase());
    }
    catch (...) {
        /* Maybe the OS doesn't allow this name */
    }
}
