/**
  *  \file ui/util.cc
  *  \brief UI Utilities
  *
  *  Graphics primitives, rather specific to our UI.
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <vector>
#include "ui/util.h"
#include "gfx/palette.h"
#include "util/strtab.h"
#include "cpluslib/utf8.h"

/** Tile area with pixmap.
    \param can    canvas to use
    \param r      area to tile with pixmap
    \param pix    pixmap to use, may be null
    \param color  when /pix/ is null, the image is filled with this color
    \param alter  X coordinate alteration. With alteration 0, the area is
                  tiled with a regular grid, like on a checkered paper.
                  With nonzero alteration, the second row is shifted
                  that many pixels to the left, the next one is shifted
                  to the right again, etc, to make the pattern look more
                  interesting. */
void
drawTiledArea(GfxCanvas& can, GfxRect r, Ptr<GfxPixmap> pix,
              uint32 color, int alter)
{
    /* fallback method. */
    if (!pix.ptr()) {
        can.drawBar(r, color, GFX_TRANSPARENT, SOLID_FILL, GFX_ALPHA_OPAQUE);
        return;
    }
    pix->tile(can, r, alter);
}

/** Draw "up" frame. Like drawRectangle(), but the rectangle is colored
    to give the effect of a 3D raised panel. */
void
drawFrameUp(GfxCanvas& can, GfxRect r)
{
    int x2 = r.x + r.w - 1;
    int y2 = r.y + r.h - 1;

    can.drawHLine(GfxPoint(r.x, r.y),   x2-r.x, standard_colors[COLOR_WHITE], SOLID_LINE, GFX_ALPHA_OPAQUE);
    can.drawVLine(GfxPoint(r.x, r.y+1), y2-r.y, standard_colors[COLOR_WHITE], SOLID_LINE, GFX_ALPHA_OPAQUE);
    can.drawHLine(GfxPoint(r.x, y2),    x2-r.x, standard_colors[COLOR_BLACK], SOLID_LINE, GFX_ALPHA_OPAQUE);
    can.drawVLine(GfxPoint(x2,  r.y), y2+1-r.y, standard_colors[COLOR_BLACK], SOLID_LINE, GFX_ALPHA_OPAQUE);
}

/** Draw "down" frame. Like drawRectangle(), but the rectangle is colored
    to give the effect of a 3D lowered panel. */
void
drawFrameDown(GfxCanvas& can, GfxRect r)
{
    int x2 = r.x + r.w - 1;
    int y2 = r.y + r.h - 1;

    can.drawHLine(GfxPoint(r.x, r.y),   x2-r.x, standard_colors[COLOR_BLACK], SOLID_LINE, GFX_ALPHA_OPAQUE);
    can.drawVLine(GfxPoint(r.x, r.y+1), y2-r.y, standard_colors[COLOR_BLACK], SOLID_LINE, GFX_ALPHA_OPAQUE);
    can.drawHLine(GfxPoint(r.x, y2),    x2-r.x, standard_colors[COLOR_WHITE], SOLID_LINE, GFX_ALPHA_OPAQUE);
    can.drawVLine(GfxPoint(x2,  r.y), y2+1-r.y, standard_colors[COLOR_WHITE], SOLID_LINE, GFX_ALPHA_OPAQUE);
}

/** Draws a dashed frame. The frame alternates between c1 and c2 in
    4-pixel segments, starting with a specified phase. Varying the
    phase causes the frame to "run around" the rectangle.
    \param can    Canvas
    \param r      Rectangle; frame will be contained in rectangle
    \param phase  Initial phase
    \param c1,c2  Colors */
void
drawDashedFrame(GfxCanvas& can, const GfxRect r, const uint8 phase, const uint32 c1, const uint32 c2)
{
    uint8 p = phase & 7;

    const int x0 = r.x;
    const int x1 = r.x + r.w - 1;
    const int y0 = r.y;
    const int y1 = r.y + r.h - 1;

    int x = x0, y = y0;
    /* top-left to top-right */
    while (x < x1) {
        can.drawPixel(GfxPoint(x, y), ((p & 7) < 4) ? c1 : c2, GFX_ALPHA_OPAQUE);
        ++p;
        ++x;
    }
    /* top-right to bottom-right */
    while (y < y1) {
        can.drawPixel(GfxPoint(x, y), ((p & 7) < 4) ? c1 : c2, GFX_ALPHA_OPAQUE);
        ++p;
        ++y;
    }
    /* bottom-right to bottom-left */
    while (x > x0) {
        can.drawPixel(GfxPoint(x, y), ((p & 7) < 4) ? c1 : c2, GFX_ALPHA_OPAQUE);
        ++p;
        --x;
    }
    /* bottom-left to top-left */
    while (y > y0) {
        can.drawPixel(GfxPoint(x, y), ((p & 7) < 4) ? c1 : c2, GFX_ALPHA_OPAQUE);
        ++p;
        --y;
    }
}

/** Split text into lines.
    \param font      Font to use
    \param width     Available width in pixels
    \param s         Text. Can contain '\n' for forced linefeeds.
    \param lines     (out) Text split into lines
    \post lines.size() != 0 */
static void
splitText(GfxFont& font, int width, const string_t s, std::vector<string_t>& lines)
{
    string_t line;
    int lastSplitPoint = 0;

    Utf8Reader rdr(s.data(), s.size());
    while (rdr.hasMore()) {
        uint32_t c = rdr.get();
        if (c == '\n') {
            /* new line */
            lines.push_back(line);
            lastSplitPoint = 0;
            line.erase();
        } else if (c == ' ') {
            /* add a word */
            line += ' ';
            lastSplitPoint = line.length();
            if (font.getTextWidth(line) >= width) {
                /* need to split here */
                lines.push_back(line.substr(0, line.length()-1));
                lastSplitPoint = 0;
                line.erase();
            }
        } else {
            /* add a character */
            string_t::size_type oldSize = line.size();
            appendUtf8(line, c);
            if (font.getTextWidth(line) > width) {
                if (!lastSplitPoint)
                    lastSplitPoint = oldSize;
                if (!lastSplitPoint)
                    lastSplitPoint = 1;
                lines.push_back(line.substr(0, lastSplitPoint));
                line.erase(0, lastSplitPoint);
                lastSplitPoint = 0;
            }
        }
    }
    /* add terminating partial line. */
    if (!line.empty() || lines.empty())
        lines.push_back(line);
}

/** Count number of lines occupied by text.
    \param font     Font to use
    \param width    Available width
    \param s        Text. Can contain '\n' */
int
countLines(GfxFont& font, int width, string_t s)
{
    std::vector<string_t> lines;
    splitText(font, width, s, lines);
    return lines.size();
}

/** Get text width occupied by word-wrapped text. Word-wraps the text
    according to the specified width, and returns the effective total
    width of the text (which may be less than \c width due to word wrap).
    \param font  Font to use
    \param width Available width
    \param s     Text. Can contain '\n' */
int
getTextWidthFormatted(GfxFont& font, int width, string_t s)
{
    std::vector<string_t> lines;
    splitText(font, width, s, lines);
    int min_width = font.getTextWidth(lines[0]);
    for (unsigned i = 1; i < lines.size(); ++i) {
        int this_width = font.getTextWidth(lines[i]);
        if (this_width > min_width)
            min_width = this_width;
    }
    return min_width;
}

/** Output text, with word wrap.
    \param ctx     Graphics context, providing font, canvas, color, and alignment
    \param x,y     Upper-left coordinate
    \param width   Maximum width
    \param s       Text. Can contain '\n' */
void
outTextFormatted(GfxContext& ctx, int x, int y, int width, string_t s)
{
    std::vector<string_t> lines;
    splitText(ctx.getFont(), width, s, lines);
    for (unsigned i = 0; i < lines.size(); ++i) {
        outText(ctx, x, y, lines[i]);
        y += ctx.getFont().getTextHeight(lines[i]);
    }
}

/** Output text, with word warp. This limits the output to a specified
    region resp. expands it to fill that region.
    \param ctx     Graphics context, providing font, canvas, color, and alignment
    \param x,y     Upper-left coordinates
    \param width   Available text width. Text is word-wrapped to this
                   width, and space to the right filled.
    \param num_lines Available number of lines. Text is limited to this
                   width, and trailing lines cleared.
    \param s       Text to display */
void
outTextFormattedF(GfxContext& ctx, int x, int y, int width, int num_lines, string_t s)
{
    std::vector<string_t> lines;
    splitText(ctx.getFont(), width, s, lines);

    for (int i = 0; i < int(lines.size()) && i < num_lines; ++i) {
        outTextF(ctx, x, y, width, lines[i]);
        y += ctx.getFont().getTextHeight(lines[i]);
    }

    if (!ctx.isTransparentBackground() && int(lines.size()) < num_lines) {
        drawBackground(ctx,
                       GfxRect(x - width * ctx.getTextAlign().x/2,
                               y,
                               width,
                               (num_lines - lines.size()) * ctx.getFont().getTextHeight("Tp")));
    }
}

/** Compute width of a text field containing an item from a StringTable. */
int
getTextWidth(GfxFont& font, const StringTable& table)
{
    int max = 0;
    for (int i = table.getFirstIndex(); i < table.getLastIndex(); i = table.getNextIndex(i)) {
        int thisWidth = font.getTextWidth(table(i));
        if (thisWidth > max)
            max = thisWidth;
    }
    return max;
}

void
getTabbedTextWidth(GfxFont& font, const StringTable& table, int& left_width, int& total_width)
{
    left_width = 0;
    total_width = 0;
    for (int i = table.getFirstIndex(); i < table.getLastIndex(); i = table.getNextIndex(i)) {
        string_t text = table(i);
        string_t::size_type n = text.find('\t');
        int thisTotal, thisLeft;
        if (n == string_t::npos) {
            thisLeft = 0;
            thisTotal = font.getTextWidth(text);
        } else {
            thisLeft = font.getTextWidth(text.substr(0, n)) + 5;
            thisTotal = thisLeft + font.getTextWidth(text.substr(n+1));
        }
        if (thisTotal > total_width) {
            total_width = thisTotal;
        }
        if (thisLeft > left_width) {
            left_width = thisLeft;
        }
    }
}
