/**
  *  \file util/nls.h
  *  \brief Definitions for National Language Support
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_UTIL_NLS_H
#define PCC_V2_UTIL_NLS_H

#include <vector>
#include <map>
#include <cstddef>
#include "cpluslib/string.h"
// #include <string>

#define _(x)  translate(x)
#define N_(x) (x)

class Stream;

class StringPool {
    /** Minimum size allocated for a pool entry. This is the size we
        allocate from the system allocator. */
    static const std::size_t MIN_POOL = 8192;

    /** Pointer to first entry of current pool. */
    char* current_pool;

    /** Number of bytes free in current pool. */
    std::size_t current_free;

    /** A single pool. */
    struct Pool {
        char*       ptr;
        std::size_t size;
        Pool()
            : ptr(0), size(0) { }
    };

    /** All allocated pools. */
    std::vector<Pool> pools;

    StringPool(const StringPool&);
    StringPool& operator=(const StringPool&);
 public:
    StringPool();
    ~StringPool();
    char* allocate(std::size_t size);
    void  clear();
};

class StringTranslator {
    StringPool strings;

    struct Compare {
        bool operator()(const char* l, const char* r) const;
    };

    /** Highest allowed indexed entry. This limit is there to
        prevent stupid overloads of the system. */
    static const int MAX_NUMBERED_ENTRY = 1000;
    
    /** Translation data. Mapping original -> translated data. */
    std::map<const char*, const char*, Compare> data;

    /** Numbered entries. */
    std::vector<const char*> numbered_data;

    StringTranslator(const StringTranslator&);
    StringTranslator& operator=(const StringTranslator&);

 public:
    StringTranslator();
    ~StringTranslator();

    const char* translate(const char* in) const;
    const char* getTranslation(const char* in) const;
    const char* getNumberedEntry(int n) const;
    void        addTranslation(const char* ori, const char* translated);
    bool        addNumberedEntry(int n, const char* text);

    void        loadLanguageFile(Stream& s);

    void        clear();

    static StringTranslator* getTranslation(string_t name);
    static bool              getCreateTranslation(string_t name, StringTranslator*& ptr);
};

StringTranslator* loadLanguageDatabase(const char* domain, const char* fn);
const char* translate(const char* in);

string_t getLanguageCode();

void initNLS();

#endif
